"use strict";
/* LICENSE BEGIN
    This file is part of the SixtyFPS Project -- https://sixtyfps.io
    Copyright (c) 2020 Olivier Goffart <olivier.goffart@sixtyfps.io>
    Copyright (c) 2020 Simon Hausmann <simon.hausmann@sixtyfps.io>

    SPDX-License-Identifier: GPL-3.0-only
    This file is also available under commercial licensing terms.
    Please contact info@sixtyfps.io for more information.
LICENSE END */
// Load the native library with `process.dlopen` instead of with `require`.
// This is only done for autotest that do not require nom or neon_cli to
// copy the lib to its right place
/**
 * @hidden
 */
function load_native_lib() {
    const os = require('os');
    process.dlopen(module, process.env.SIXTYFPS_NODE_NATIVE_LIB, os.constants.dlopen.RTLD_NOW);
    return module.exports;
}
/**
 * @hidden
 */
let native = !process.env.SIXTYFPS_NODE_NATIVE_LIB ? require('../native/index.node') : load_native_lib();
/**
 * @hidden
 */
class Component {
    constructor(comp) {
        this.comp = comp;
    }
    show() {
        this.comp.show();
    }
    send_mouse_click(x, y) {
        this.comp.send_mouse_click(x, y);
    }
    send_keyboard_string_sequence(s) {
        this.comp.send_keyboard_string_sequence(s);
    }
}
require.extensions['.60'] =
    function (module, filename) {
        var c = native.load(filename);
        module.exports[c.name()] = function (init_properties) {
            let comp = c.create(init_properties);
            let ret = new Component(comp);
            c.properties().forEach((x) => {
                Object.defineProperty(ret, x, {
                    get() { return comp.get_property(x); },
                    set(newValue) { comp.set_property(x, newValue); },
                    enumerable: true,
                });
            });
            c.signals().forEach((x) => {
                Object.defineProperty(ret, x, {
                    get() {
                        let signal = function () { return comp.emit_signal(x, [...arguments]); };
                        signal.setHandler = function (callback) { comp.connect_signal(x, callback); };
                        return signal;
                    },
                    enumerable: true,
                });
            });
            return ret;
        };
    };
/**
 * @hidden
 */
class NullPeer {
    rowDataChanged(row) { }
    rowAdded(row, count) { }
    rowRemoved(row, count) { }
}
/**
 * ArrayModel wraps a JavaScript array for use in `.60` views. The underlying
 * array can be modified with the [[ArrayModel.push]] and [[ArrayModel.remove]] methods.
 */
class ArrayModel {
    /**
     * Creates a new ArrayModel.
     *
     * @param arr
     */
    constructor(arr) {
        this.a = arr;
        this.notify = new NullPeer();
    }
    rowCount() {
        return this.a.length;
    }
    rowData(row) {
        return this.a[row];
    }
    setRowData(row, data) {
        this.a[row] = data;
        this.notify.rowDataChanged(row);
    }
    /**
     * Pushes new values to the array that's backing the model and notifies
     * the run-time about the added rows.
     * @param values
     */
    push(...values) {
        let size = this.a.length;
        Array.prototype.push.apply(this.a, values);
        this.notify.rowAdded(size, arguments.length);
    }
    // FIXME: should this be named splice and hav ethe splice api?
    /**
     * Removes the specified number of element from the array that's backing
     * the model, starting at the specified index. This is equivalent to calling
     * Array.slice() on the array and notifying the run-time about the removed
     * rows.
     * @param index
     * @param size
     */
    remove(index, size) {
        let r = this.a.splice(index, size);
        this.notify.rowRemoved(index, size);
    }
    get length() {
        return this.a.length;
    }
    values() {
        return this.a.values();
    }
    entries() {
        return this.a.entries();
    }
}
module.exports = {
    private_api: native,
    ArrayModel: ArrayModel,
    Timer: {
        singleShot: native.singleshot_timer,
    }
};
